"""
Simple way to implement counters via sqlite

sample usage:
python3 -m dataset_utils.counters increment_cli counters.sqlite completed

"""
import dataset
from dataset_utils.sqlite import connect, get_or_create, ConnectMode
import fire
import time
from typing import Union


def get_or_create_counters(db: Union[str, dataset.Database], table_name: str):
    return get_or_create(
        db, table_name,
        columns=['epoch', 'value'],
        types={'epoch': 'float', 'value': 'integer'},
        index=['epoch']
    )


class SqliteCounters(object):
    def __init__(self, db, mode=ConnectMode.JOURNAL):
        if isinstance(db, str):
            self.db = connect(db, mode=mode)
        else:
            assert isinstance(db, dataset.Database)
            self.db = db
        self.table_cache = {}

    def increment(self, counter, value=1):
        if counter not in self.table_cache:
            _, table = get_or_create_counters(self.db, counter)
            self.table_cache[counter] = table
        table = self.table_cache[counter]
        table.insert({'ts': time.time(), 'value': value})

    def window_sum(self, counter, lookback):
        ts = time.time() - lookback
        out = self.db.query(f'select sum(value) from {counter} where ts >= {ts};')
        row = next(out)
        sum_val = row['sum(value)']
        return sum_val

    def prune(self):
        # TODO - implement later
        pass


def increment(dbpath: str, table_name: str, value=1):
    sc = SqliteCounters(dbpath)
    sc.increment(table_name, value=value)


def window_sum(db: str, table_name: str, lookback):
    sc = SqliteCounters(db, mode=ConnectMode.READ_ONLY)
    sum_val = sc.window_sum(table_name, lookback)
    if sum_val is None:
        return 0
    return sum_val


def window_sum_gte(db: str, table_name: str, lookback, value):
    sum_val = window_sum(db, table_name, lookback)
    assert sum_val >= value


if __name__ == '__main__':
    fire.Fire()



"""
Key-Value

Oops...TODO - copy in the features from dict_cache
"""
import json

from typing import Any, Dict, Tuple, Set, Union, List
from .sqlite import get_or_create, ConnectMode
from .buffered_writer import BufferedTableWriter, ConflictChecker

class KVLite(object):

    @classmethod
    def resolve_type(cls, elem_type):
        if elem_type == 'json':
            return 'string'
        return elem_type

    def __init__(self, path, key_name='key', value_name='value', key_type='string', value_type='string', table_name='items', connect_mode=ConnectMode.JOURNAL):
        self.key_type = key_type
        self.value_type = value_type
        self.key_name = key_name
        self.value_name = value_name
        self.db, self.table = get_or_create(
            path, table_name, mode=connect_mode, primary_id=self.key_name, columns=[self.value_name],
            types={self.key_name: self.resolve_type(self.key_type), self.value_name: self.resolve_type(self.value_type)}
        )

    @classmethod
    def is_type(cls, elem_type, elem):
        if elem_type == 'string':
            return isinstance(elem, str)
        else:
            raise Exception('Unknown type: ' + elem_type)

    @classmethod
    def serialize_type(cls, elem_type, elem):
        if elem_type in ('string', 'integer', 'float'):
            return elem
        elif elem_type == 'json':
            return json.dumps(elem)
        else:
            raise Exception('Do not know how to serialize type: ' + elem_type)

    @classmethod
    def deserialize_type(cls, elem_type, elem):
        if elem_type in ('string', 'integer', 'float'):
            assert cls.is_type(elem_type, elem)
            return elem
        elif elem_type == 'json':
            return json.loads(elem)
        else:
            raise Exception('Do not know how to serialize type: ' + elem_type)

    def _find_elem(self, key_):
        return self.table.find_one(**{self.key_name: key_})

    def get(self, key_, default=None):
        elem = self._find_elem(key_)
        if elem is None:
            return default
        tmp_val = elem[self.value_name]
        return self.deserialize_type(self.value_type, tmp_val)

    # TODO - implement def items(...)

    def __getitem__(self, key_):
        val = self.get(key_)
        if val is None:
            raise KeyError('Error no key in db: ' + key_)
        return val

    def __contains__(self, key_):
        return self._find_elem(key_) is not None

    def __setitem__(self, key_, value):
        #assert self.is_type(self.key_type, key_)
        #assert self.is_type(self.value_type, value)
        self.table.upsert({
            # TODO - for now, we are not serializing the key
            self.key_name: key_,
            self.value_name: self.serialize_type(self.value_type, value)
        }, [self.key_name])

    def set_multiple(self, d: Dict[str, Dict[str, Any]]):
        b = BufferedTableWriter(self.table, key_column=self.key_name, key_conflicts=ConflictChecker.OVERWRITE)
        for k, v in d.items():
            assert isinstance(k, str)
            b.insert({self.key_name: k, self.value_name: self.serialize_type(self.value_type, v)})
        b.force_flush()

    def get_multiple(self, keys: Union[List[str], str]) -> Tuple[Dict[str, Any], Set[str]]:
        if isinstance(keys, str):
            keys = [keys]
        assert isinstance(keys, list)
        keys_set = set(keys)
        assert len(keys) == len(keys_set), 'Error: there are duplicate keys in your request'
        results = {}
        for e in self.table.find(key=keys):
            k, v = e[self.key_name], e[self.value_name]
            assert isinstance(k, str)
            results[k] = self.deserialize_type(self.value_type, v)
        remaining = keys_set - set(results.keys())
        assert len(results) + len(remaining) == len(keys)
        return results, remaining




from typing import List, Union, Dict, Tuple, Any, Set
from dataset_utils.sqlite import get_or_create
from .buffered_writer import BufferedTableWriter, ConflictChecker
import os


class DictCache(object):
    def __init__(self, dbpath, table_name='default', key_column='key', value_column='value'):
        # model id is model_name + model version
        assert os.path.abspath(dbpath) == dbpath, f'error: please pass in an absolute path: {dbpath} vs {os.path.abspath(dbpath)}'
        self.dbpath = dbpath
        self.table_name = table_name
        self.key_column = key_column
        self.value_column = value_column
        self.db, self.table = get_or_create(
            self.dbpath, self.table_name, primary_id=self.key_column, columns=[self.value_column],
            types={self.key_column: 'string', self.value_column: 'string'})

    def set_multiple(self, d: Dict[str, Dict[str, str]]):
        b = BufferedTableWriter(self.table, key_column=self.key_column, key_conflicts=ConflictChecker.OVERWRITE)
        for k, v in d.items():
            assert isinstance(k, str)
            assert isinstance(v, str)
            b.insert({self.key_column: k, self.value_column: v})
        b.force_flush()

    def get_multiple(self, keys: Union[List[str], str]) -> Tuple[Dict[str, str], Set[str]]:
        if isinstance(keys, str):
            keys = [keys]
        assert isinstance(keys, list)
        keys_set = set(keys)
        assert len(keys) == len(keys_set), 'Error: there are duplicate keys in your request'
        results = {}
        for e in self.table.find(key=keys):
            k, v = e[self.key_column], e[self.value_column]
            assert isinstance(k, str)
            assert isinstance(v, str)
            results[k] = v
        remaining = keys_set - set(results.keys())
        assert len(results) + len(remaining) == len(keys)
        return results, remaining




import os
# TODO - should finish this sometime
from typing import List, Union, Dict, Tuple, Any, Set
from dataset_utils.sqlite import get_or_create, ConnectMode, connect
from dataset_utils.buffered_writer import BufferedTableWriter, ConflictChecker
import os
import dataset
from collections import defaultdict
from pypika import Table, Query


def chunks(lst, n):
    # https://stackoverflow.com/a/312464
    """Yield successive n-sized chunks from lst."""
    for i in range(0, len(lst), n):
        yield lst[i:i + n]


class ScoreCache(object):
    def __init__(self, dbpath, table_name: str, rel_from: str):
        # model id is model_name + model version
        assert os.path.abspath(dbpath) == dbpath, f'error: please pass in an absolute path: {dbpath} vs {os.path.abspath(dbpath)}'
        dne = not os.path.exists(dbpath)
        self.dbpath = dbpath
        self.table_name = table_name
        self.rel_from = rel_from
        self.db, self.table = get_or_create(
            self.dbpath, self.table_name, primary_id='path', types={'path': 'string'},
            mode=ConnectMode.WAL)
        # assert self.table_name in self.db

    def create_indicies(self):
        assert isinstance(self.table, dataset.Table)
        for col in self.table.columns:
            # NOTE: the below code only creates the index if it doesn't already exist
            self.table.create_index([col])

    def set_multiple(self, d: Dict[str, Dict[str, float]]):
        b = BufferedTableWriter(self.table, key_column='path', key_conflicts=ConflictChecker.OVERWRITE)
        for path, scores in d.items():
            assert isinstance(scores, dict), f'error scores is not a dict: {str(d)}'
            assert isinstance(path, str), f'path is not a string: {str(path)}'
            assert os.path.isabs(path), f'error: {path} is not an abs path'
            row: Dict[str, Any] = scores.copy()
            max_all = 0.0
            max_per = defaultdict(float)
            for label, score in scores.items():
                assert isinstance(score, float), f'Error contains non-floats: {str(scores)}'
                max_all = max(max_all, score)
                label_prefix = 'max_' + label.split('_')[0]
                max_per[label_prefix] = max(max_per[label_prefix], score)

            for label, score in max_per.items():
                row[label] = score

            row['max_all'] = max_all
            relpath = os.path.relpath(path, self.rel_from)
            row['path'] = relpath
            row['parent_folder'] = relpath.split('/')[0]
            b.insert(row)
        b.force_flush()
        self.create_indicies()

    def labels(self):
        result = []
        for col in self.table.columns:
            if col in ('parent_folder', 'path'):
                continue
            result.append(col)
        return result

    def count(self, label, threshold):
        params = {
            'label': label,
            'threshold': {
                '>=': threshold
            }
        }
        return self.table.count(**params)

    def top_predictions(
            self, target_class: str, any_path_prefix=None, none_path_prefix=None, folders=None, not_folders=None,
            order_by='default', order='desc', page=0, limit=10000,
            score_filter=None
    ) -> List[Dict[str, Any]]:
        if order_by in ('default', None):
            order_by = target_class
        if isinstance(folders, str):
            folders = [folders]
        assert order in ('desc', 'asc')
        results = []
        offset = page * limit
        import logging
        logging.error(f'order_by: {order_by}')
        logging.error(f'limit: {limit}')
        logging.error(f'offset: {offset}')
        from pypika import Table, Query, Order, Criterion
        ctable = Table(self.table_name)
        q = Query.from_(ctable).select(ctable.star)
        assert folders or not_folders, 'at least one needs to be set'
        assert not (folders and not_folders), 'folders and not_folders cannot be set'
        if folders:
            q = q.where(ctable.parent_folder.isin(folders))
        else:
            assert not_folders
            q = q.where(ctable.parent_folder.notin(not_folders))

        if score_filter is not None:
            isinstance(score_filter, list)
            for e in score_filter:
                assert len(e) == 3
                col, op, val = e
                if op == '>=':
                    q = q.where(getattr(ctable, col).gte(val))
                elif op == '<=':
                    q = q.where(getattr(ctable, col).lte(val))
                else:
                    raise Exception(f'Unknown operator: {op}')

        if any_path_prefix:
            if isinstance(any_path_prefix, str):
                any_path_prefix = [any_path_prefix]
            assert isinstance(any_path_prefix, list)
            criterions = []
            for path_prefix in any_path_prefix:
                assert isinstance(path_prefix, str)
                # either of the two criterions are equivalent...not sure which one is faster
                #q = q.where(ctable.path.glob(path_prefix + '*'))
                criterions.append(ctable.path.glob(path_prefix + '*'))
                # criterions.append(tag_table.path.glob(f'{folder}/*'))
            q = q.where(Criterion.any(criterions))

        ###### START MONKEY PATCH #####
        # until this PR gets merged:
        # https://github.com/kayak/pypika/pull/586/files

        from pypika.terms import Term, BasicCriterion, Comparator

        class Matching(Comparator):
            not_like = " NOT LIKE "
            like = " LIKE "
            not_ilike = " NOT ILIKE "
            ilike = " ILIKE "
            regex = " REGEX "
            bin_regex = " REGEX BINARY "
            as_of = " AS OF "
            glob = " GLOB "
            not_glob = " NOT GLOB "

        def not_glob(local_self, expr: str):
            return BasicCriterion(Matching.not_glob, local_self, Term.wrap_constant(expr))
        Term.not_glob = not_glob

        ###### END MONKEY PATCH #####

        if none_path_prefix:
            if isinstance(none_path_prefix, str):
                none_path_prefix = [none_path_prefix]
            criterions = []
            for path_prefix in none_path_prefix:
                assert isinstance(path_prefix, str)
                criterions.append(ctable.path.not_glob(path_prefix + '*'))
            q = q.where(Criterion.all(criterions))

        q = q.orderby(getattr(ctable, order_by), order=getattr(Order, order))
        # after the score, order by the path
        q = q.orderby(ctable.path)
        q = q.offset(offset)
        q = q.limit(limit)
        logging.error('should run: ' + str(q))
        #for row in self.table.find(parent_folder={'in': folders}, order_by=order_by, _offset=offset, _limit=limit):
        for row in self.db.query(str(q)):
            row['path'] = os.path.join(self.rel_from, row['path'])
            results.append(row)
        return results

    def table_counts(self):
        tmp_results = self.db.query(f"SELECT name FROM sqlite_master WHERE type='table' AND name='{self.table_name}';")
        existing_tables = set()
        for row in tmp_results:
            existing_tables.add(row['name'])
        return existing_tables

    def get_multiple(self, paths: Union[List[str], str], default_score=None, order_by=None, limit=None) -> Tuple[Dict[str, Dict[str, Any]], List[str]]:
        existing_tables = self.table_counts()
        if self.table_name not in existing_tables:
            return {}, paths
        if isinstance(paths, str):
            paths = [paths]
        for path in paths:
            assert os.path.isabs(path)
        # we convert to relpath because we use it for everything
        relpaths = [os.path.relpath(path, self.rel_from) for path in paths]
        #import logging
        #logging.error('new paths: ' + str(paths))
        assert isinstance(relpaths, list)
        keys_set = set(paths)
        assert len(relpaths) == len(keys_set), 'Error: there are duplicate keys in your request'
        results = {}
        score_table = Table(self.table_name)

        for relpaths_chunk in chunks(relpaths, 10000):
            # we chunk the reads b/c there's a limit for how big a query can be
            q = Query.from_(score_table).select(
                score_table.star,
            )
            q = q.where(score_table.path.isin(relpaths_chunk))
            if order_by:
                q = q.orderby(order_by)
            q = q.limit(limit)
            for row in self.db.query(str(q)):
                relpath = row.pop('path')
                # row['path'] = relpath
                path = os.path.join(self.rel_from, relpath)
                parent_folder = row.pop('parent_folder')
                for label, score in row.items():
                    if score is None and default_score is not None:
                        score = default_score
                    assert isinstance(score, float), f'Error contains non-floats: {str(row)}'
                assert isinstance(row, dict)
                row['parent_folder'] = parent_folder
                row['path'] = relpath
                results[path] = row
        remaining = keys_set - set(results.keys())
        remaining = [os.path.join(self.rel_from, p) for p in remaining]
        assert len(results) + len(remaining) == len(relpaths)
        return results, remaining




import dataset
from dataset.types import Types
from enum import Enum
from urllib.parse import urlparse
from sqlalchemy import event
import sqlite3
import os
from dataset.database import Database
from typing import Tuple


class ConnectMode(Enum):
    # Rollback journal (the default mode)
    JOURNAL = 1
    # Write-Ahead-Log:
    # https://www.sqlite.org/wal.html
    WAL = 2
    READ_ONLY = 3


def connect(path, mode=ConnectMode.JOURNAL, ensure_exists=False):
    # NOTE: if you open in read-only mode, we always check to see it exists
    # otherwise, if ensure_exists is False, we might create it if it doesn't exist
    if ensure_exists:
        assert os.path.exists(path)
    if mode == ConnectMode.JOURNAL:
        return connect_journal(path)
    elif mode == ConnectMode.WAL:
        return connect_wal(path)
    else:
        assert mode == ConnectMode.READ_ONLY
        return connect_ro(path)


def connect_journal(path):
    return dataset.connect('sqlite:///' + path, sqlite_wal_mode=False)


def connect_ro(path):
    # https://github.com/pudo/dataset/issues/136
    # https://stackoverflow.com/questions/27910829/sqlalchemy-and-sqlite-shared-cache
    # https://stackoverflow.com/questions/48218065/programmingerror-sqlite-objects-created-in-a-thread-can-only-be-used-in-that-sa
    # See "URI Connections":
    # https://docs.sqlalchemy.org/en/14/dialects/sqlite.html
    assert os.path.exists(path)
    creator = lambda: sqlite3.connect('file:' + path + '?mode=ro&check_same_thread=false', uri=True)
    return dataset.connect('sqlite:///', engine_kwargs={'creator': creator}, sqlite_wal_mode=False)


def connect_wal(path) -> dataset.Database:
    assert not path.startswith('sqlite:')
    url = 'sqlite:///' + path
    db = dataset.connect(url, sqlite_wal_mode=True)
    return db


def get_or_create(path_or_db, table_name, mode=ConnectMode.JOURNAL, primary_id=None, columns=None, types=None, index=None) -> Tuple[dataset.Database, dataset.Table]:
    # by default, types are all string (text or string - they're the same in sqlite)
    # more readable than calling get_or_create_from_dict
    generated_dict = {}
    if columns is not None:
        assert isinstance(columns, list) or isinstance(columns, tuple) or isinstance(columns, dict)

    if isinstance(columns, dict):
        assert types is None
        types = columns
        columns = list(columns.keys())
        try:
            if primary_id:
                columns.remove(primary_id)
        except ValueError:
            pass
    if types is None:
        types = dict()

    def get_type(column, default='text'):
        type_ = types.get(column)
        if type_ is None:
            type_ = default
        if isinstance(type_, str):
            return getattr(dataset.types.Types, type_)
        assert not isinstance(type_, str)
        return type_

    if primary_id:
        generated_dict['primary'] = (primary_id, get_type(primary_id))
    generated_dict['columns'] = []
    if columns:
        for column in columns:
            generated_dict['columns'].append((column, get_type(column)))
    if index:
        generated_dict['index'] = index

    db = get_or_create_from_dict(path_or_db, {table_name: generated_dict}, mode=mode)
    table = db.load_table(table_name)
    return db, table


def get_or_create_from_dict(path_or_db, table_info: dict, mode=ConnectMode.JOURNAL) -> dataset.Database:
    if not isinstance(path_or_db, dataset.Database):
        assert isinstance(path_or_db, str)
        db = connect(path_or_db, mode)
    else:
        db = path_or_db

    def get_type(type_):
        if isinstance(type_, str):
            return getattr(db.types, type_)
        return type_

    for table_name in table_info.keys():
        ct = table_info[table_name]
        if table_name not in db:
            # create column if it doesn't exist
            # Create table_name
            #assert 'primary' in ct
            #assert 'columns' in ct
            if 'primary' in ct:
                primary_id, primary_key = ct['primary']
                table = db.create_table(table_name, primary_id=primary_id, primary_type=get_type(primary_key))
            else:
                table = db.create_table(table_name)
        else:
            table = db[table_name]
        assert isinstance(table, dataset.Table)
        assert isinstance(ct['columns'], list)
        for col in ct['columns']:
            col_name, col_type = col
            if not table.has_column(col_name):
                table.create_column(col_name, get_type(col_type))
        if 'index' in ct:
            for cc in ct['index']:
                idxcols = cc
                if not isinstance(cc, list):
                    idxcols = [cc]
                if not table.has_index(idxcols):
                    table.create_index(idxcols)
        # NOTE: we have to do this to ensure the table is created by the time we return,
        #       esp for WAL mode
        #tmp_table = db.load_table(table_name)
        # even this line below
        #assert len(tmp_table.columns) >= 0
        # assert table_name in db
    return db


class RamCache():
    # TODO - in the middle of implementing

    @classmethod
    def from_path(cls, path):
        pass

    # For now, read-only
    def __init__(self, path, cache_dir='/dev/shm/'):
        assert os.path.isdir(cache_dir)
        assert os.path.exists(path)
        import shutil
        shutil.copy(path, cache_dir)
        from pathlib import Path
        self.cached_path = os.path.join(cache_dir, Path(path).name)
        assert os.path.isfile(self.cached_path)
        self.db = connect_ro(self.cached_path)
        self.closed = False

    def close(self):
        self.closed = True
        import shutil
        os.remove(self.cached_path)

